/* 
 * Glippy
 * Copyright © 2010 Wojciech Kowalczyk
 * The program is distributed under the terms of the GNU General Public License Version 3.
 * See LICENCE for details.
 */
namespace Glippy
{
	using Gdk;
	using Gtk;
	using System.Collections.Generic;

	/// <summary>
	/// Clipboard class. Reads system clipboard and stores data.
	/// </summary>
	public class GlippyClipboard
	{		
		#region Variables, properties & delegates
		private GlippyConfig _GConfig;
		private Clipboard _ClipboardPaste, _ClipboardMouse;
		private ClipboardTextReceivedFunc _TextReceivedHandler, _MouseReceivedHandler;
		private ClipboardImageReceivedFunc _ImageReceivedHandler;				
		private string _MouseContent = "";
		private bool _MouseHandled = true;								
		private List<GlippyItem> _StoredItems;
		
		public List<GlippyItem> StoredItems			
		{ get { return _StoredItems; }	}		
		
		public GlippyItem ClipboardContent
		{ 
			set
			{				
				if (value.Image != null)
					_ClipboardPaste.Image = value.Image;
				else
					_ClipboardPaste.Text = value.Text;
			}
		}		
		
		public string ClipboardMouseContent
		{ set { _ClipboardMouse.Text = value; } }
						
		public delegate void ClipboardChangeHandler(object sender, ChangedArgs args);
		public event ClipboardChangeHandler ClipboardChanged;
		#endregion
		
		public GlippyClipboard (GlippyConfig gconfig)
		{			
			// Creating GConf client
			this._GConfig = gconfig;			
			this._StoredItems = new List<GlippyItem>();			
			
			// Adding handlers to delegates
			this._ImageReceivedHandler += ImageReceived;
			this._TextReceivedHandler += TextReceived;
			this._MouseReceivedHandler += MouseReceived;								
			
			// Getting clipboards and adding ownerchange handlers
			_ClipboardPaste = Clipboard.Get (Gdk.Selection.Clipboard);
			_ClipboardPaste.OwnerChange += OnClipboardOwnerChange;								
			_ClipboardMouse = Clipboard.Get(Gdk.Selection.Primary);								
			_ClipboardMouse.OwnerChange += OnMouseClipboardOwnerChange;
		}
				
		#region Ownerchange event handlers
		// FIXME Remove handlers instead of checking bool values
		private void OnClipboardOwnerChange(object sender, OwnerChangeArgs args)
		{	
			if (!_GConfig.EnableClipboard)
				return;
			#if DEBUG
				System.Console.WriteLine("Handling keyboard clipboard.");
			#endif
			if (_ClipboardPaste.WaitIsTextAvailable())							
				_ClipboardPaste.RequestText(_TextReceivedHandler);			
			else if (_ClipboardPaste.WaitIsImageAvailable() && _GConfig.EnableImagesClipboard)				
				_ClipboardPaste.RequestImage(_ImageReceivedHandler);																		
		}
				
		private void OnMouseClipboardOwnerChange(object sender, OwnerChangeArgs args)
		{			
			if (!_GConfig.EnableMouseClipboard)
				return;
			#if DEBUG
				System.Console.WriteLine("Handling mouse clipboard.");
			#endif
			if (_ClipboardMouse.WaitIsTextAvailable())
					_ClipboardMouse.RequestText(_MouseReceivedHandler);
		}
		#endregion
		
		// Received from clipboard text handler
		private void TextReceived(Clipboard clip, string text)
		{						
			if (text == "" || text == null)
				return;
														
			#if DEBUG						
				System.Console.WriteLine("Type: text.");
			#endif
			foreach (GlippyItem i in _StoredItems)			
				if (text == i.Text)			
				{
					#if DEBUG						
						System.Console.WriteLine("\"{0}\" exists, moving item up.", text);
					#endif
					_StoredItems.Remove(i);
					_StoredItems.Insert(0, i);
					if (ClipboardChanged != null)
						ClipboardChanged (this, null);			
					return;				
				}
			
			AddStoredItem(text, null);
			RemoveRedundantItems();
						
			if (_GConfig.Synchronize && _GConfig.EnableMouseClipboard)
				_ClipboardMouse.Text = text;
			
			if (ClipboardChanged != null)
				ClipboardChanged (this, null);			
		}
		
		// Received from clipboard image handler
		private void ImageReceived(Clipboard clip, Pixbuf pixbuf)
		{		
			string text;
			#if DEBUG						
				System.Console.WriteLine("Type: image.");
			#endif
			foreach (GlippyItem i in _StoredItems)			
				if (i.Image != null && pixbuf != null)					
					if (ComparePixbufs(pixbuf, i.Image))	// compare pixel by pixel, equals doesn't work probably because both images have different pointers
					{												
						#if DEBUG						
							System.Console.WriteLine("Image exists, moving up.");
						#endif	
						pixbuf.Dispose();
						_StoredItems.Remove(i);
						_StoredItems.Insert(0, i);
						if (ClipboardChanged != null)
							ClipboardChanged (this, null);			
						return;
					}				
			
			System.Text.StringBuilder textbuilder = new System.Text.StringBuilder();				
			textbuilder.Append("[");
			textbuilder.Append(Mono.Unix.Catalog.GetString("Image"));
			textbuilder.Append(" ");
			textbuilder.Append(pixbuf.Width);
			textbuilder.Append("x");
			textbuilder.Append(pixbuf.Height);
			textbuilder.Append(" ");
			textbuilder.Append(System.DateTime.Now.ToString());
			textbuilder.Append("]");
			text = textbuilder.ToString();		
			
			AddStoredItem(text, pixbuf);
			RemoveRedundantItems();
			
			if (ClipboardChanged != null)
				ClipboardChanged (this, null);			
		}
		
		// Received from mouse clipboard text handler
		private void MouseReceived(Clipboard clip, string text)
		{
			#if DEBUG						
				System.Console.WriteLine("Type: text.");
			#endif	
			if (_MouseHandled)
			{
				#if DEBUG						
					System.Console.WriteLine("Rehandle blocked.");
				#endif	
				_MouseHandled = false;
				return;
			}								
			
			if (text != null)
			{					
				if (text == _MouseContent)
					return;
				
				foreach (GlippyItem i in _StoredItems)			
					if (text == i.Text)			
					{
						#if DEBUG						
							System.Console.WriteLine("\"{0}\" exists, moving up.", text);
						#endif	
						_StoredItems.RemoveAt(0);
						_StoredItems.Remove(i);
						_StoredItems.Insert(0, i);
						if (ClipboardChanged != null)
							ClipboardChanged (this, null);			
						return;				
					}
				
				AddStoredItem(text, null);
				
				// It's sooo huuuge... Checks which way current text is changing, if none, we're in begining
				// FIXME optimize
				if (!(text.StartsWith(_MouseContent) && text.Length > _MouseContent.Length) && !(_MouseContent.StartsWith(text) && text.Length < _MouseContent.Length) && !(text.EndsWith(_MouseContent) && text.Length > _MouseContent.Length) && !(_MouseContent.EndsWith(text) && text.Length < _MouseContent.Length))
				{
					if ((text.Contains(_MouseContent) && text.Length > _MouseContent.Length) || (text.Length == 1 && _MouseContent.Length == 1 && text != _MouseContent))
					{
						#if DEBUG						
							System.Console.WriteLine("\"{0}\" is growing.", text);
						#endif	
						StoredItems.RemoveAt(1);
					}
				}
				else
					if (StoredItems.Count > 1)
					{
						#if DEBUG						
							System.Console.WriteLine("\"{0}\" is growing.", text);
						#endif	
						StoredItems.RemoveAt(1);
					}
				
				RemoveRedundantItems();
				_MouseContent = text;					
				_MouseHandled = true;
				
				
				if (_GConfig.Synchronize && _GConfig.EnableClipboard)
					_ClipboardPaste.Text = _MouseContent;
				
				if (ClipboardChanged != null)
					ClipboardChanged (this, null);			
			}											
		}
		
		// Adds new item to list
		private void AddStoredItem(string text, Gdk.Pixbuf pixbuf)
		{
			System.Text.StringBuilder label = new System.Text.StringBuilder();
			label.Append(text);
			
			// Cuts part of text, because we don't want too long menu				
			if (label.Length > _GConfig.ItemLength)
			{
				label.Remove(_GConfig.ItemLength/2 - 1, label.Length - _GConfig.ItemLength);
				label.Insert(label.Length/2, "...");
			}
			
			// Removing new lines in label
			if (_GConfig.RemoveNewLines)				
				label.Replace(System.Environment.NewLine, " ");				
						
			GlippyItem item = new GlippyItem(label.ToString(), text, pixbuf);											
			_StoredItems.Insert (0, item);			
			#if DEBUG						
				System.Console.WriteLine("Item \"{0}\" added with label: \"{1}\".", item.Text, item.Label);
			#endif	
		}
		
		// Removes redundant items
		private void RemoveRedundantItems()
		{
			if (_StoredItems.Count > _GConfig.ClipboardSize)			
				for (int i = _StoredItems.Count - 1; i >= _GConfig.ClipboardSize; i--)
				{
					if (_StoredItems[i].Image != null)
					{
						_StoredItems[i].Image.Dispose();					
						_StoredItems[i].Image = null;
					}
					#if DEBUG						
						System.Console.WriteLine("Removing \"{0}\"", _StoredItems[i].Text);
					#endif	
					_StoredItems.RemoveAt(i);
				}							
		}
		
		// Clears clipboard
		public void Clear()
		{						
			foreach (GlippyItem i in _StoredItems)							
				if (i.Image != null)
				{
					i.Image.Dispose();
					i.Image = null;
				}			
			_StoredItems.Clear();	
			_ClipboardPaste.Clear();
			_ClipboardMouse.Clear();
			#if DEBUG						
				System.Console.WriteLine("Cliboard clean.");
			#endif	
		}						
				
		// Compares pixbuffs pixel by pixel 
		public static bool ComparePixbufs(Gdk.Pixbuf a, Gdk.Pixbuf b)
		{							
			if ((a == null || b == null) || (a.Width != b.Width || a.Height != b.Height))
			{
				#if DEBUG						
					System.Console.WriteLine("Images have different sizes, false.");
				#endif	
				return false;			
			}									
			
			unsafe
			{
				byte* ptra = (byte*)a.Pixels.ToPointer();
				byte* ptrb = (byte*)b.Pixels.ToPointer();
				long size = a.Width * a.Height * 3;
				for (int i = 0; i < size; i++)
				{
					if (*ptra != *ptrb)
					{
						#if DEBUG						
							System.Console.WriteLine("Images have different sizes, false.");
						#endif	
						return false;
					}
					ptra++;
					ptrb++;
				}					
			}
						
			#if DEBUG						
				System.Console.WriteLine("Images are the same, true.");
			#endif	
			return true;
		}		
	}		
}
