/* 
 * Glippy
 * Copyright © 2010 Wojciech Kowalczyk
 * The program is distributed under the terms of the GNU General Public License Version 3.
 * See LICENCE for details.
 */
namespace Glippy
{
	using Gtk;
	using Mono.Unix;
	using Mono.Unix.Native;
	using System;

	/// <summary>
	/// Glippy's core class, creates config, clipboard, hotkey, SIGTERM-wait thread, base for icon-like classes.
	/// </summary>
	public class GlippyCore
	{
		protected Menu _Menu;
		protected GlippyClipboard _GClip;
		protected GlippyConfig _GConfig;
		protected System.Threading.Thread _SignalThread;
		protected GlippyXHotkey _Hotkey;
		protected GlippySettings _SettingsWindow;
		
		public GlippyCore()
		{
			_GConfig = new GlippyConfig();					
			_GClip = new GlippyClipboard (_GConfig);			
			_GClip.ClipboardChanged += OnChange;
					
			// Initialize hotkeys
			_Hotkey = new GlippyXHotkey(_GConfig.HotKeyModifier, _GConfig.HotkeyKeycode);				
			if (_GConfig.EnableHotkeys)			
				_Hotkey.GrabKey();	
			
			_Hotkey.HotkeyPressed += ShowMenu;
			
			if (_GConfig.SaveClipboardToFile)
				GlippyHistory.ReadHistory(_GClip.StoredItems, _GConfig);
			
			_Menu = new Menu();
			SetMenu();															
			#if DEBUG						
				Console.WriteLine("Core running fine.");
			#endif	
		}
		
		// Let's create menu!
		protected virtual void SetMenu()
		{
			MenuItem menuitem;			
						
			// Mr. Muscle POWAH!
			foreach (Widget w in _Menu.Children)
			{
				#if DEBUG						
					Console.WriteLine("{0} item removed.", w.ToString());
				#endif	
				_Menu.Remove(w);
				w.Destroy();
				w.Dispose();								
			}
			
			// Load items
			if (_GClip.StoredItems.Count != 0)
			{								
				foreach (GlippyItem i in _GClip.StoredItems)
				{
					if (i.Text == null || i.Text == "")
					{
						_GClip.StoredItems.Remove(i);
						break;
					}					
					menuitem = new MenuItem(i.Label);					
					menuitem.Activated += OnItemActivated;						
					
					_Menu.Add(menuitem);					
					#if DEBUG						
						Console.WriteLine("\"{0}\" item added.", i.Label);
					#endif	
				}							
			}
			
			#region Default buttons
			// Separator
			_Menu.Add(new SeparatorMenuItem());
			
			// Clear button
			MenuItem menuItemClear = new MenuItem(Catalog.GetString("_Clear clipboard"));
			menuItemClear.Activated += delegate(object sender, EventArgs e)
			{ _GClip.Clear(); OnChange(this, null); };
			_Menu.Add(menuItemClear);
					
			// Settings button
			MenuItem menuItemSettings = new MenuItem(Catalog.GetString("_Settings"));
			menuItemSettings.Activated += delegate(object sender, EventArgs e)
			{
				if (_SettingsWindow != null)
				{
					_SettingsWindow.Destroy();
					_SettingsWindow.Dispose();
				}
				_SettingsWindow = new GlippySettings(_GConfig, _Hotkey);
				_SettingsWindow.SettingsChanged += OnChange;
				_SettingsWindow.ShowAll();
			};
			_Menu.Add(menuItemSettings);
		
			// About button
			MenuItem menuItemAbout = new MenuItem(Catalog.GetString("About"));
			menuItemAbout.Activated += delegate(object sender, EventArgs e) { GlippyAbout.Show(); };
			_Menu.Add(menuItemAbout);
			
			// Quit button
			MenuItem menuItemQuit = new MenuItem (Catalog.GetString("_Quit"));			
			menuItemQuit.Activated += delegate { Syscall.kill(Syscall.getpid(), Signum.SIGTERM); };	 // OMG so ugly...		
			_Menu.Add(menuItemQuit);						
			#endregion
			
			_Menu.ShowAll();			
			#if DEBUG						
				Console.WriteLine("Menu built.");
			#endif	
		}
	
		// Move selected item top and change clipboard content
		protected virtual void OnItemActivated(object sender, EventArgs e)
		{ 
			int index = Array.IndexOf(_Menu.Children, sender);				
						
			GlippyItem item = _GClip.StoredItems[index];
			#if DEBUG						
				Console.WriteLine("{0} activated, moving up and setting as clipboard's content", item.Label);
			#endif
			_GClip.StoredItems.Remove(item);
			_GClip.StoredItems.Insert(0, item);									
			_GClip.ClipboardContent = item;			
			_GClip.ClipboardMouseContent = item.Text;
			
			OnChange(this, null);				
		}
		
		// Clipboard and settings change handler
		protected virtual void OnChange(object sender, ChangedArgs e)
		{			
		}			
		
		// Refreshes and shows menu
		protected virtual void ShowMenu(object sender, EventArgs e)
		{			
			#if DEBUG						
				Console.WriteLine("Menu showed.");
			#endif	
			SetMenu();
			_Menu.Popup();
		}			
		
		// SIGTERM handler in another thread			
		public void HandleSignal()
		{			
			_SignalThread = new System.Threading.Thread(delegate () {
				#if DEBUG						
					Console.WriteLine("SIGTERM & SIGINT handler started.");
				#endif	
				UnixSignal[] signals = new UnixSignal[] { new UnixSignal(Signum.SIGTERM), new UnixSignal(Signum.SIGINT) };
				UnixSignal.WaitAny(signals);	
				Application.Invoke(delegate {
					OnSaveYourself(this, null);
					OnDie(this, null);
				});		
			});
			_SignalThread.Start();
		}
		
		// Saves history before loging off
		public void OnSaveYourself(object sender, Gnome.SaveYourselfArgs args)
		{
			#if DEBUG
				Console.WriteLine("Saving history.");
			#endif
			if (_GConfig.SaveClipboardToFile)
				GlippyHistory.WriteHistory(_GClip.StoredItems);								
		}
						
		// Cleans up and quits
		public void OnDie(object sender, EventArgs args)
		{						
			#if DEBUG						
				Console.WriteLine("SIGTERM or SIGINT received, bye bye.");
			#endif	
			_SignalThread.Abort();			
			Application.Quit();				
		}
	}
}
