/* 
 * Glippy
 * Copyright © 2010 Wojciech Kowalczyk
 * The program is distributed under the terms of the GNU General Public License Version 3.
 * See LICENCE for details.
 */
namespace Glippy
{
	using Gdk;
	using System;
	using System.Runtime.InteropServices;

	/// <summary>
	///  Hotkeys support. Provides key and keyboard grabbing, emits HotkeyPressed.
	/// </summary>
	public class GlippyXHotkey
	{	
		#region Variables, properties & delegates
		private const int _KEYPRESS = 2;
	    private const int _GRABMODEASYNC = 1;
		private const string _ARCH_X86_64 = "x86_64";
		private string _Architecture;		
	    private ModifierType _Modifiers;
	    private int _KeyCode;
		private Window _RootWin;
		private IntPtr _XDisplay;
		private IntPtr _XWindow;
		private FilterFunc _Filter;		
		private bool _Grabbed;
		private bool _ReGrab;		
		public delegate void GlippyXHotkeyHandler(object sender, Gtk.ChangedArgs args);
    	public event GlippyXHotkeyHandler HotkeyPressed;		
		public event GlippyXHotkeyHandler KeyGrabbed;
		public delegate int XErrorHandler(IntPtr a, IntPtr b);		
		private XErrorHandler _ErrorHandler;
		
		public ModifierType Modifiers
		{ get { return _Modifiers; } }
		public Key Key
		{ get { return (Key)XKeycodeToKeysym(_XDisplay, _KeyCode, 0); } }		
		public int Keycode
		{ get { return _KeyCode; } }
		#endregion
	
		public GlippyXHotkey(ModifierType modifier, int keycode)
		{
			MainClass.GetArchitecture(out _Architecture);			
			#if DEBUG
				Console.WriteLine("Architecture is {0}", _Architecture);
			#endif
														
			GetXPointers();
			
			_Modifiers = modifier;
			_KeyCode = keycode;
			_Grabbed = false;
			
			_ErrorHandler = new XErrorHandler(XKeyAlreadyInUse);
			IntPtr errhandler = Marshal.GetFunctionPointerForDelegate(_ErrorHandler);
			XSetErrorHandler(errhandler);
			#if DEBUG
				Console.WriteLine("XErrorHandler set.");
			#endif
		}
		
		// Grabs key defined as keycode in _KeyCode entry
	    public void GrabKey()
	    {				 			
			if (_Grabbed)
				return;		
						
			#if DEBUG
				Console.WriteLine("Grabbing key {0} {1}.", _Modifiers.ToString(), _KeyCode);
			#endif
			uint modifiers = (uint)_Modifiers;						
			XGrabKey(_XDisplay, _KeyCode, modifiers, _XWindow, false, _GRABMODEASYNC, _GRABMODEASYNC);   					
			modifiers += (uint) ModifierType.LockMask;
			XGrabKey(_XDisplay, _KeyCode, modifiers, _XWindow, false, _GRABMODEASYNC, _GRABMODEASYNC);     								
			modifiers += (uint) ModifierType.Mod2Mask;
			XGrabKey(_XDisplay, _KeyCode, modifiers, _XWindow, false, _GRABMODEASYNC, _GRABMODEASYNC);     											
			modifiers -= (uint) ModifierType.LockMask;
			XGrabKey(_XDisplay, _KeyCode, modifiers, _XWindow, false, _GRABMODEASYNC, _GRABMODEASYNC);     								
			_Filter += FilterKeyCatchFunction;
			_RootWin.AddFilter(_Filter);	
			#if DEBUG
				Console.WriteLine("Window filter added.");
			#endif
			_Grabbed = true;							
	    }

		// Ungrabs key
	    public void UngrabKey()
	    {			
			if (!_Grabbed)
				return;
						
			#if DEBUG
				Console.WriteLine("Ungrabbing key {0} {1}.", _Modifiers.ToString(), _KeyCode);
			#endif
			uint modifiers = (uint)_Modifiers;
			XUngrabKey(_XDisplay, _KeyCode, modifiers, _XWindow);						
			modifiers += (uint) ModifierType.LockMask;
			XUngrabKey(_XDisplay, _KeyCode, modifiers, _XWindow);						
			modifiers += (uint) ModifierType.Mod2Mask;
			XUngrabKey(_XDisplay, _KeyCode, modifiers, _XWindow);						
			modifiers -= (uint) ModifierType.LockMask;
			XUngrabKey(_XDisplay, _KeyCode, modifiers, _XWindow);						
			_RootWin.RemoveFilter(_Filter);
			_Filter -= FilterKeyCatchFunction;	 
			#if DEBUG
				Console.WriteLine("Window filter removed.");
			#endif
			_Grabbed = false;						
	    }
		
		// Grabs whole keyboard
		public void GrabKeyboard()
		{
			_ReGrab = _Grabbed;
			#if DEBUG
				Console.WriteLine("Grabbing keyboard, regrab: {0}.", _ReGrab.ToString());
			#endif
			if (_Grabbed)
				UngrabKey();					
			Keyboard.Grab(_RootWin, false, 0);					
			_Filter += FilterGetKeyFunction;
			_RootWin.AddFilter(_Filter);	
			#if DEBUG
				Console.WriteLine("Window filter added.");
			#endif
		}
				
		// Ungrabs keyboard
		private void UngrabKeyboard()
		{		
			#if DEBUG
				Console.WriteLine("Ungrabbing keyboard, regrab: {0}.", _ReGrab.ToString());
			#endif
			Keyboard.Ungrab(0);
			_RootWin.RemoveFilter(_Filter);						
			_Filter -= FilterGetKeyFunction;			
			#if DEBUG
				Console.WriteLine("Window filter removed.");
			#endif
			
			if (_ReGrab)
				GrabKey();
						
			if (KeyGrabbed != null)
				KeyGrabbed(this, null);
		}
		
		// Catches key combination pressed, when keyboard was grabbed
		private FilterReturn FilterGetKeyFunction(IntPtr xEvent, Event evnt)
		{				
			XKeyEvent32 xKeyEvent32;
			XKeyEvent64 xKeyEvent64;
			Key key;
			ModifierType modifier = ModifierType.None;
			
			if (_Architecture == _ARCH_X86_64)
			{								
				xKeyEvent64 = (XKeyEvent64)Marshal.PtrToStructure(xEvent, typeof(XKeyEvent64));
				key = (Key)XKeycodeToKeysym(_XDisplay, (int)xKeyEvent64.keycode, 0);
				modifier = (ModifierType)xKeyEvent64.state;
				#if DEBUG
					Console.WriteLine("Got XSERVER64 event, {0} {1}", modifier.ToString(), key.ToString());
				#endif
			}
			else
			{
				xKeyEvent32 = (XKeyEvent32)Marshal.PtrToStructure(xEvent, typeof(XKeyEvent32));
				key = (Key)XKeycodeToKeysym(_XDisplay, (int)xKeyEvent32.keycode, 0);
				modifier = (ModifierType)xKeyEvent32.state;
				#if DEBUG
					Console.WriteLine("Got XSERVER32 event, {0} {1}", modifier.ToString(), key.ToString());
				#endif
			}
				
			if (key == Key.Escape || key == Key.BackSpace)
			{
				#if DEBUG
					Console.WriteLine("{0} key, fall back.", key.ToString());
				#endif
				UngrabKeyboard();
				return FilterReturn.Remove;
			}
			
			// C-C-C-COMBO BREAKER!!111oneoneoneoenneo
			if (modifier == ModifierType.None || modifier == ModifierType.LockMask || modifier == ModifierType.Mod2Mask || modifier == ModifierType.Mod5Mask)
				return FilterReturn.Continue;											
			if (key == Key.Control_L || key == Key.Control_R || key == Key.Alt_L || key == Key.Alt_R || key == Key.Super_L || key == Key.Super_R || key == Key.Shift_L || key == Key.Shift_R)
				return FilterReturn.Continue;			
									
			// Clean caps and num locks
			if (modifier.ToString().Contains(ModifierType.LockMask.ToString()))			
				modifier = (ModifierType)((int)modifier - (int)ModifierType.LockMask);
			if (modifier.ToString().Contains(ModifierType.Mod2Mask.ToString()))
				modifier = (ModifierType)((int)modifier - (int)ModifierType.Mod2Mask);
						
			_Modifiers = modifier;
			_KeyCode = XKeysymToKeycode(_XDisplay, (int)key);
			
			#if DEBUG
				Console.WriteLine("Everything fine, grabbed key {0} {1}", modifier.ToString(), key.ToString());
			#endif
			
			UngrabKeyboard();
			return FilterReturn.Remove;
		}
		
		// Catches specific key combination and emits HotkeyPressed
	    	private FilterReturn FilterKeyCatchFunction(IntPtr xEvent, Event evnt)
	    	{	
			XKeyEvent32 xKeyEvent32;
			XKeyEvent64 xKeyEvent64;
			int keycode;
			ModifierType modifier = ModifierType.None;
			int type;
			
			if (_Architecture == _ARCH_X86_64)
			{
				xKeyEvent64 = (XKeyEvent64)Marshal.PtrToStructure(xEvent, typeof(XKeyEvent64));
				keycode = (int)xKeyEvent64.keycode;
				modifier = (ModifierType)xKeyEvent64.state;
				type = xKeyEvent64.type;
				#if DEBUG
					Console.WriteLine("Got XSERVER64 event, {0} {1}", modifier.ToString(), keycode.ToString());
				#endif
			}
			else
			{
				xKeyEvent32 = (XKeyEvent32)Marshal.PtrToStructure(xEvent, typeof(XKeyEvent32));
				keycode = (int)xKeyEvent32.keycode;
				modifier = (ModifierType)xKeyEvent32.state;
				type = xKeyEvent32.type;
				#if DEBUG
					Console.WriteLine("Got XSERVER32 event, {0} {1}", modifier.ToString(), keycode.ToString());
				#endif
			}
					
			// Clean caps and num locks
			if (modifier.ToString().Contains(ModifierType.LockMask.ToString()))			
				modifier = (ModifierType)((int)modifier - (int)ModifierType.LockMask);
			if (modifier.ToString().Contains(ModifierType.Mod2Mask.ToString()))
				modifier = (ModifierType)((int)modifier - (int)ModifierType.Mod2Mask);
			
			if (type == _KEYPRESS)	        
		            if (keycode == _KeyCode && modifier == _Modifiers)
				if (HotkeyPressed != null)
				{
					#if DEBUG
						Console.WriteLine("Emitting HotKeyPressed.");
					#endif
		           		HotkeyPressed(this, null);
				}
					
	        	return FilterReturn.Remove;						
		}
	    	
				
		#region XLib		
		private void GetXPointers()
		{
			_RootWin = Global.DefaultRootWindow;
			_XWindow = GetXWindow();
			_XDisplay = GetXDisplay(_RootWin);
			#if DEBUG
				Console.WriteLine("Got unmanaged pointers.");
			#endif
		}
		
		// Whatever, just don't kill Glippy, you bastard! :<
		private int XKeyAlreadyInUse(IntPtr display, IntPtr evnt)
		{			
			#if DEBUG
				Console.WriteLine("X failure.");
			#endif
			_Modifiers = (ModifierType)12;
			_KeyCode = 54;								
										
			if (KeyGrabbed != null)
				KeyGrabbed(this, new Gtk.ChangedArgs());
			
			return -1;
		}
		
		// Getters
	    private static IntPtr GetXWindow()
	    {			
			return gdk_x11_get_default_root_xwindow();
		}	

		private static IntPtr GetXDisplay(Window win)
		{
			return gdk_x11_display_get_xdisplay(win.Display.Handle);
		}
					
		// Xlib and libgtk imports for global hotkeys, I'm sure there is a simpler solution, but who cares?
	    [DllImport("libgtk-x11-2.0.so.0")] private static extern IntPtr gdk_x11_display_get_xdisplay (IntPtr gdk_display);
		[DllImport("libgtk-x11-2.0.so.0")] private static extern IntPtr gdk_x11_get_default_root_xwindow();
	    [DllImport("libX11")] private static extern int XKeysymToKeycode(IntPtr display, int key);
		[DllImport("libX11")] private static extern int XKeycodeToKeysym(IntPtr display, int key, int index);
	    [DllImport("libX11")] private static extern int XGrabKey(IntPtr display, int keycode, uint modifiers, IntPtr grab_window, bool owner_events, int pointer_mode, int keyboard_mode);
	    [DllImport("libX11")] private static extern int XUngrabKey(IntPtr display, int keycode, uint modifiers, IntPtr grab_window);
		[DllImport("libX11")] private static extern int XSetErrorHandler(IntPtr handler);
				
		// FIXME X server has different types on different archs; temporary workaround		
		[StructLayout(LayoutKind.Sequential)]
	    internal struct XKeyEvent64
	    {
	        public int type;
	        public ulong serial;
	        public int send_event;
	        public IntPtr display;			
			public ulong window;
	        public ulong root;
	        public ulong subwindow;
	        public ulong time;					
	        public int x, y;
	        public int x_root, y_root;
	        public uint state;
	        public uint keycode;
	        public int same_screen;
	    
		}	
		
		[StructLayout(LayoutKind.Sequential)]
	    internal struct XKeyEvent32
	    {
	        public int type;
	        public uint serial;
	        public int send_event;
	        public IntPtr display;
	        public uint window;
	        public uint root;
	        public uint subwindow;
	        public uint time;
	        public int x, y;
	        public int x_root, y_root;
	        public uint state;
	        public uint keycode;
	        public int same_screen;
	    }
		#endregion		
	}
}

